/* 

 Basic battery tester V1.6 - Gives runtime for a battery with a fixed load which can be compared with known good battery (or specifications on data sheet) 
 to determine battery condition and health. Also gives approximate amp hour reading; this tester uses a resistive load, accurate amp hour ratings are not
 really possible without a constant current load tester. Final rating will be around +/- 5% of actual capacity.
 
 Version 1.6 - modified to test 12V SLA batteries. Due to voltage multiplied by approximately 11 voltage reading accuracy is +/- 8mV approx in 4mV steps. 
			   accuracy cannot really be increased further without using a better ADAC / lower voltage reference IC (say 2.048 volts) and / or multisampling. 
			   Added check to measure offload voltage before running test. If too low advises user to check to see if battery is charged (or its totally knackered)
   
 Version 1.5 - better display layout. Does not show voltage and current display if battery voltage rebounds after load disconnect.
 
 Version 1.4 - now uses TL431 as voltage reference (the TL431 has 1% accuracy of 2.5V reference) actual measured voltages are within 10mv now
  
*/

#include <LiquidCrystal.h>
#include <avr/pgmspace.h>

const float     aRefVoltage = 2.5; // actual measured 2.493V - near enough. Does vary with temperature but within 1% of 2.5V. 

//   The integrator works by accumulating the sampled voltage values from the start until
//   hitting a pre-determined low-voltage threshold. Since they are sampled at a 
//   known period, the number of samples taken cancels out and the accumulation of all samples 
//   simply needs to be scaled by a factor
//
//   Let's make some definitions to show the derivation of how this is so:
//   I = current 
//   V = load voltage
//   Rl = load resistance
//   samples = number of voltage samples made
//   sensor = value read out of analog A->D
//   rate = number of samples taken per second
//   
//                                        ave(V)
//   capacity = ∫I ~= ave(I) * time = ----------- * time
//                                          Rl
//
//   Where:
//                ∑(V)     ∑(sensor) * aRefVoltage/1024
//     ave(V) = -------  = ---------------------------------
//              samples              samples
//
//      time  =  samples/rate 
//
//   Thus:
//               ∑(sensor) * aRefVoltage        samples
//   capacity = ----------------------------- * ----------
//                samples * 1024 * Rl            rate
//
//               ∑(sensor) * aRefVoltage 
//   capacity = ------------------------------ = ∑(sensor) * quanta 
//                  1024 * Rl * rate
//
//
//                aRefVoltage
//   quanta =  --------------------  
//               1024 * Rl * rate
//
//   rate in this sketch is 1 Hz (1/sec).  
//   Thus, the units of quanta are: volt*seconds/(Ω*samples) = amp*seconds per sample
//
//   What we really want are mA*Hours, so, scale by 1hr/3600seconds and 1000mA/A, or simply 1/3.6
//
//   So, finally:
//   quanta = aRefVoltage/(Rl*1024*3.6)  in units of mA*hours per sample

const double quanta = 0.00022757; // 2.5/(2.78*1024*3.6) 10985.472 - actually 2.98 ohms including fuseholder etc. Measured with voltcraft low resistance meter.


// Define low voltage threshold where any remaining capacity is "unusable"
//        lowThreshold = 10.8 / VDIV * 1024/adacRef
const int lowThreshold = 399;

int sensorPin          = A1;   // select the analog input pin for the voltage measurement
int ledPin             = 13;   // select the pin for the LED
int speakerPin         = 7;    // speaker / buzzer pin
int fetGatePin         = 8;    // select the pin for the FET
int sensorValue        = 0;    // unscaled sensor output 
float voltage          = 0;    // measured voltage // the AVR has no floating point instruction so floating point math takes 10000+ CPU cycles (slow!)
float loadcurrent      = 0;    // calculated load current 
double mAH             = 0;    // calculated mAh
long accumulator       = 0;    // sum of all unscaled sensor values sampled
int epoch              = 0;    // seconds since battery was inserted 
int lowVolts           = 0;    // debounce the low voltage threshold
bool done              = false;// Voltage has dropped below threshold
bool batteryIn         = false;// Battery present

#define R1 2.78; // change this to your load resistors actual value. Also don't forget to re-calculate the quanta as well!
#define VDIV 11.124 // ADAC is only 10 bit so voltage readings are in 5mv steps. 11.074 - changed to 11.124 to compensate for voltage drop along cable and fuseholder
#define VCAL 1.171 // as it is a resistive load Ah rating will be slightly less than actual capacity, this compensates for that. 
#define START_BEEP        2500      // frequency in Hz
#define DONE_BEEP         1800

int  delayInMillis = 0;
int  idle = 0;
unsigned long previousMillis = 0; // last time update
long interval = 60000; // interval at which to do something (milliseconds)

// initialize the the LCD library with the numbers of the interface pins
//  LCD Pins  --  RS, EN, D4, D5, D6, D7
LiquidCrystal lcd(12, 11,  5,  4,  3,  2);

void sendtoSerial()
{

//send to serial port

Serial.print(voltage); 
Serial.print(",");
Serial.print(epoch /60 % 60); // print current minute of elapsed time since start of test

} 

void setup() {
  // declare the ledPin and fetGatePin as an OUTPUT:
  pinMode(ledPin,     OUTPUT);  
  pinMode(fetGatePin, OUTPUT);
  pinMode(speakerPin, OUTPUT);  

  // set analog reference to external
  analogReference(EXTERNAL);
  Serial.begin(9600);

  // set up the LCD's number of rows and columns: 
  lcd.begin(16, 2);

  //lcd.print(F("2019 A Smith"));
  //lcd.setCursor(0, 1);
  lcd.print(F("Ver 1.6"));
  delay(2000); // leave software version number on screen for 2 seconds
  lcd.setCursor(0, 0);
  lcd.clear();
  lcd.print(F("12V Lead Acid")); 
  lcd.setCursor(0, 1);
  lcd.print(F("Battery Tester"));
  delay(5000); // leave intro message on display for 5 seconds
  lcd.clear();  
  lcd.setCursor(0, 0);
  sensorValue = analogRead(sensorPin);

  
  //Print a helpful start-up message to the LCD. Comment out if powering from battery under test.
  
    //lcd.clear();
    //lcd.print(F("Connect batt."));
    //lcd.setCursor(0, 1);
    //lcd.print(F("to start test"));  
  
  if (sensorValue < 450) // battery must be 12.2V or above. Was 443 
  {
	  lcd.clear();
	  lcd.print("Not charged");
	  lcd.setCursor(0, 1);
	  lcd.print("or defective!");
    tone(speakerPin, DONE_BEEP,999);  // Play long beep for 999 miliseconds
	  delay(10000); // wait 10 seconds to ensure user has seen that battery needs charging or is defective.
  }
	  
	  else 
	  {
		  lcd.clear();
		  lcd.print("Voltage OK!");
		  lcd.setCursor(0, 1);
		  lcd.print("Test starting..");
		  delay(2000); // let user know battery charge is OK.
	  }
  
  Serial.begin(9600); 
}

void loop() {
  
  sensorValue = analogRead(sensorPin);    

  if (!batteryIn && (sensorValue > 436)) {
    // only start if voltage is over 11.8V. Prevents test from restarting after load disconnection.
    batteryIn   = true;
    done        = false;
    epoch       = 0;
    accumulator = 0;
    lowVolts    = 0;
    digitalWrite(ledPin, LOW);
    digitalWrite(fetGatePin, HIGH);
    // clear out when LCD when starting over
    lcd.clear();
    tone(speakerPin, START_BEEP,50);   // short beep
    
  }
  else if (batteryIn && done && (sensorValue <= 20)) {
    // consider the battery removed only after finishing the last measurement to 
    // debounce a glitchy contact
    batteryIn=false; 
  } 
  else if (batteryIn && !done) {
    // Running state during discharge state

 unsigned long currentMillis = millis();

  if(currentMillis - previousMillis > interval) 
  {
     previousMillis = currentMillis;  

    //sendtoSerial(); // for diagnostics
    
  }
    
    voltage = sensorValue*aRefVoltage/1024.0 *VDIV;
    loadcurrent = voltage/R1+0.15;
  
    // print voltage
    lcd.setCursor(0, 0);
    lcd.print(voltage);
    lcd.print("V");
    if (voltage <10.00)
    lcd.print(" ");
    lcd.setCursor(7, 0);
    lcd.print(loadcurrent);
    lcd.print("A RUN");
          
    if (!done) {
      accumulator += sensorValue;
      // print mAH
      mAH = accumulator*quanta;
     
      lcd.setCursor(0,1);
      lcd.print(double(mAH /100 *VCAL));
      lcd.print("Ah ");
      //if  (mAH < 10000) // shift display if more than 10,000mAh (10Ah)
      if  (mAH /100 *VCAL > 9.99) // shift display if more than 10Ah 
      lcd.setCursor(8,1); 
      else
      lcd.setCursor(7,1);
      
      int chours, cminutes, cseconds;
      chours = epoch/60 / 60;
      cminutes = epoch /60 % 60;
      cseconds = epoch %60;
      if (chours <= 9)
      {
      lcd.print("0");
      }
      lcd.print(chours);
      lcd.print(":");
      if (cminutes <= 9)
      {
       lcd.print("0");
      }
       lcd.print(cminutes);
       lcd.print(":");
      if (cseconds <= 9)
      {
       lcd.print("0");
      }
       lcd.print(cseconds);          
    }
        
    if (!done) {
  
      // lowVolts requires 10 seconds in the last 20 before being done
      if (sensorValue < lowThreshold) { lowVolts++;}
      else if (lowVolts > 0)          { lowVolts--;}  
  
      // If it's below threshold for 10 of 20 samples, bail out
      if (lowVolts > 10) {
        done=true;
        tone(speakerPin, DONE_BEEP,999);  // done beep
        digitalWrite(fetGatePin, LOW);
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(F("Capacity "));
        lcd.print(double(mAH /100 *VCAL)); // display result in amp hours
        lcd.print(" Ah ");
        lcd.setCursor(0, 1);
        lcd.print(F("Runtime "));
        
        int chours, cminutes, cseconds;
        chours = epoch/60 / 60;
        cminutes = epoch /60 % 60;
        if (chours <= 9)
        {
         lcd.print("0");
        }
       lcd.print(chours);
       lcd.print("h ");
       if (cminutes <= 9)
        {
         lcd.print("0");
        }
       lcd.print(cminutes);
       lcd.print("m");        
    
      }
    }

    epoch++;
  } // batteryIn -- main routine


  
  if (done) {
    // When done, flash the LED to get attention
    digitalWrite(ledPin, HIGH);
    delay(500);
    digitalWrite(ledPin, LOW);
    delay(499);
  } else {
    // Since the processing takes some time prior to the delay, we'll assume 1mS 
    // This could stand to be improved with an interrupt routine that is kicked off
    // before all the processing starts for each loop
    delay(1000); 
  }

  
}
